# labels.py

__all__ = [
    "READ",
    "WRITE",
    "DELETE",
    "SEARCH",
    "TEXT",
    "BYTES",
    "JSON"
]

READ = 'read'
WRITE = 'write'
DELETE = 'delete'
SEARCH = 'search'

TEXT = 'text'
BYTES = 'bytes'
JSON = 'json'